import logging
from typing import List, Optional

from django_bulk_hooks.core.changeset import ChangeSet, RecordChange
from django_bulk_hooks.core.context import OperationContext
from django_bulk_hooks.factory import create_hook_instance
from django_bulk_hooks.hooks.adapters import ConditionAdapter, SignatureAdapter

logger = logging.getLogger(__name__)


class HookDispatcher:
    """
    Dispatch hooks based on conditions and priority.
    """

    def __init__(self, registry):
        self.registry = registry

    def dispatch(self, context: OperationContext, event: str) -> None:
        hooks = self._get_hooks(context.model_cls, event)
        if not hooks:
            return

        logger.info(
            "Dispatching %s for %s: %d hooks",
            event,
            context.model_cls.__name__,
            len(hooks),
        )

        for handler_cls, method_name, condition, priority in hooks:
            self._execute_hook(handler_cls, method_name, condition, context, event)

    def _get_hooks(self, model_cls: type, event: str):
        try:
            hooks = self.registry.get_hooks_including_abstract(model_cls, event)
        except Exception:
            hooks = []
        if not hooks:
            hooks = self.registry.get_hooks(model_cls, event)
        return hooks

    def _execute_hook(
        self,
        handler_cls: type,
        method_name: str,
        condition: Optional[object],
        context: OperationContext,
        event: str,
    ) -> None:
        logger.info(
            "[DISPATCHER] Evaluating hook: %s.%s for event=%s, model=%s",
            handler_cls.__name__,
            method_name,
            event,
            context.model_cls.__name__,
        )
        
        filtered_context = context
        if condition is not None:
            logger.debug(
                "[DISPATCHER] Evaluating condition: %s (total records: %d)",
                condition.__class__.__name__,
                len(context.changeset),
            )
            matching_changes = [
                change
                for change in context.changeset
                if ConditionAdapter.check(condition, change, context)
            ]
            if not matching_changes:
                logger.debug(
                    "[DISPATCHER]   → %s.%s: no matching records (condition filtered all)",
                    handler_cls.__name__,
                    method_name,
                )
                return
            logger.info(
                "[DISPATCHER]   → Condition matched %d/%d records",
                len(matching_changes),
                len(context.changeset),
            )
            filtered_context = _FilteredContext(context, matching_changes)
        else:
            logger.debug("[DISPATCHER]   → No condition, all records match")

        logger.info(
            "[DISPATCHER] Executing hook: %s.%s with %d records",
            handler_cls.__name__,
            method_name,
            len(filtered_context.changeset),
        )
        handler = create_hook_instance(handler_cls)
        method = getattr(handler, method_name)
        SignatureAdapter.call_hook(method, filtered_context)
        logger.info("[DISPATCHER] Hook completed: %s.%s", handler_cls.__name__, method_name)


class _FilteredContext:
    """
    Lightweight wrapper to expose a filtered ChangeSet.
    """

    def __init__(self, base: OperationContext, changes: List[RecordChange]):
        self._base = base
        self._changeset = ChangeSet(base, tuple(changes))

    @property
    def intent(self):
        return self._base.intent

    @property
    def changeset(self):
        return self._changeset

    @property
    def operation_type(self) -> str:
        return self._base.operation_type

    @property
    def model_cls(self) -> type:
        return self._base.model_cls

    @property
    def operation_id(self):
        return self._base.operation_id

    def is_field_in_scope(self, field_name: str) -> bool:
        return self._base.is_field_in_scope(field_name)

    @property
    def old_records_map(self):
        return self._base.old_records_map

    @property
    def new_records(self):
        return tuple(self._changeset.new_records)



"""
HookQuerySet - Django QuerySet with hook support.

This is a thin coordinator that delegates all complex logic to services.
It follows the Facade pattern, providing a simple interface over the
complex coordination required for bulk operations with hooks.
"""

import logging

from django.conf import settings
from django.db import models
from django.db import transaction

from django_bulk_hooks.helpers import extract_pks

logger = logging.getLogger(__name__)


class HookQuerySet(models.QuerySet):
    """
    QuerySet with hook support.

    This is a thin facade over BulkOperationCoordinator. It provides
    backward-compatible API for Django's QuerySet while integrating
    the full hook lifecycle.

    Key design principles:
    - Minimal logic (< 10 lines per method)
    - No business logic (delegate to coordinator)
    - No conditionals (let services handle it)
    - Transaction boundaries only
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._coordinator = None

    @classmethod
    def with_hooks(cls, queryset):
        """
        Apply hook functionality to any queryset.

        This enables hooks to work with any manager by applying hook
        capabilities at the queryset level rather than through inheritance.

        Args:
            queryset: Any Django QuerySet instance

        Returns:
            HookQuerySet instance with the same query parameters
        """
        if isinstance(queryset, cls):
            return queryset  # Already has hooks

        # Create a new HookQuerySet with the same parameters as the original queryset
        hook_qs = cls(
            model=queryset.model,
            query=queryset.query,
            using=queryset._db,
            hints=getattr(queryset, "_hints", {}),
        )

        # Preserve any additional attributes from the original queryset
        # This allows composition with other queryset enhancements
        cls._preserve_queryset_attributes(hook_qs, queryset)

        return hook_qs

    @classmethod
    def _preserve_queryset_attributes(cls, hook_qs, original_qs):
        """
        Preserve attributes from the original queryset.

        This enables composition with other queryset enhancements like
        queryable properties, annotations, etc.
        """
        # Copy non-method attributes that might be set by other managers
        for attr_name in dir(original_qs):
            if not attr_name.startswith("_") and not hasattr(cls, attr_name) and not callable(getattr(original_qs, attr_name, None)):
                try:
                    value = getattr(original_qs, attr_name)
                    setattr(hook_qs, attr_name, value)
                except (AttributeError, TypeError):
                    # Skip attributes that can't be copied
                    continue

    @property
    def coordinator(self):
        """Lazy initialization of coordinator"""
        if self._coordinator is None:
            from django_bulk_hooks.operations import BulkOperationCoordinator

            self._coordinator = BulkOperationCoordinator(self)
        return self._coordinator

    def _use_intent_architecture(self) -> bool:
        config = getattr(settings, "DJANGO_BULK_HOOKS", {})
        return bool(config.get("USE_INTENT_ARCHITECTURE", False))

    @transaction.atomic
    def bulk_create(
        self,
        objs,
        batch_size=None,
        ignore_conflicts=False,
        update_conflicts=False,
        update_fields=None,
        unique_fields=None,
        bypass_hooks=False,
    ):
        """
        Create multiple objects with hook support.

        This is the public API - delegates to coordinator.
        """
        if self._use_intent_architecture():
            return self._bulk_create_v3(
                objs=objs,
                batch_size=batch_size,
                ignore_conflicts=ignore_conflicts,
                update_conflicts=update_conflicts,
                update_fields=update_fields,
                unique_fields=unique_fields,
                bypass_hooks=bypass_hooks,
            )
        return self.coordinator.create(
            objs=objs,
            batch_size=batch_size,
            ignore_conflicts=ignore_conflicts,
            update_conflicts=update_conflicts,
            update_fields=update_fields,
            unique_fields=unique_fields,
            bypass_hooks=bypass_hooks,
        )

    @transaction.atomic
    def bulk_update(
        self,
        objs,
        fields=None,
        batch_size=None,
        bypass_hooks=False,
        **kwargs,
    ):
        """
        Update multiple objects with hook support.

        This is the public API - delegates to coordinator.

        Args:
            objs: List of model instances to update
            fields: List of field names to update (optional, will auto-detect if None)
            batch_size: Number of objects per batch
            bypass_hooks: Skip all hooks if True

        Returns:
            Number of objects updated
        """
        # DEBUG: Log incoming fields parameter
        logger.debug(f"🟦 QUERYSET.bulk_update ENTRY: fields={fields}, objs count={len(objs) if objs else 0}")

        # If fields is None, auto-detect changed fields using analyzer
        if fields is None:
            from django_bulk_hooks.operations.analyzer import ModelAnalyzer

            fields = ModelAnalyzer(self.model).detect_changed_fields(objs)
            logger.debug(f"🟦 QUERYSET.bulk_update: Auto-detected fields={fields}")
            if not fields:
                return 0

        if self._use_intent_architecture():
            return self._bulk_update_v3(
                objs=objs,
                fields=fields,
                batch_size=batch_size,
                bypass_hooks=bypass_hooks,
            )

        logger.debug(f"🟦 QUERYSET.bulk_update: Calling coordinator.update with fields={fields}")
        return self.coordinator.update(
            objs=objs,
            fields=fields,
            batch_size=batch_size,
            bypass_hooks=bypass_hooks,
        )

    @transaction.atomic
    def update(self, bypass_hooks=False, **kwargs):
        """
        Update QuerySet with hook support.

        This is the public API - delegates to coordinator.

        Args:
            bypass_hooks: Skip all hooks if True
            **kwargs: Fields to update

        Returns:
            Number of objects updated
        """
        if self._use_intent_architecture():
            return self._update_v3(update_kwargs=kwargs, bypass_hooks=bypass_hooks)
        return self.coordinator.update_queryset(
            update_kwargs=kwargs,
            bypass_hooks=bypass_hooks,
        )

    @transaction.atomic
    def bulk_delete(
        self,
        objs,
        bypass_hooks=False,
        **kwargs,
    ):
        """
        Delete multiple objects with hook support.

        This is the public API - delegates to coordinator.

        Args:
            objs: List of objects to delete
            bypass_hooks: Skip all hooks if True

        Returns:
            Tuple of (count, details dict)
        """
        if self._use_intent_architecture():
            return self._bulk_delete_v3(objs=objs, bypass_hooks=bypass_hooks)

        # Filter queryset to only these objects
        pks = extract_pks(objs)
        if not pks:
            return 0

        # Create a filtered queryset
        filtered_qs = self.filter(pk__in=pks)

        # Use coordinator with the filtered queryset
        from django_bulk_hooks.operations import BulkOperationCoordinator

        coordinator = BulkOperationCoordinator(filtered_qs)

        count, details = coordinator.delete(
            bypass_hooks=bypass_hooks,
        )

        # For bulk_delete, return just the count to match Django's behavior
        return count

    @transaction.atomic
    def delete(self, bypass_hooks=False):
        """
        Delete QuerySet with hook support.

        This is the public API - delegates to coordinator.

        Args:
            bypass_hooks: Skip all hooks if True

        Returns:
            Tuple of (count, details dict)
        """
        if self._use_intent_architecture():
            return self._delete_v3(bypass_hooks=bypass_hooks)
        return self.coordinator.delete(
            bypass_hooks=bypass_hooks,
        )

    # ==================== Intent-centric v3 API ====================

    def _bulk_create_v3(
        self,
        objs,
        batch_size=None,
        ignore_conflicts=False,
        update_conflicts=False,
        update_fields=None,
        unique_fields=None,
        bypass_hooks=False,
    ):
        if not objs:
            return objs

        from django_bulk_hooks.core.context import OperationContext
        from django_bulk_hooks.core.intent import IntentBuilder
        from django_bulk_hooks.orchestration.orchestrator import get_orchestrator

        intent_builder = (
            IntentBuilder(self.model)
            .for_create()
            .with_unique_fields(unique_fields)
            .with_options(
                batch_size=batch_size,
                ignore_conflicts=ignore_conflicts,
                update_conflicts=update_conflicts,
                bypass_hooks=bypass_hooks,
                using_db=self.db,
            )
        )
        if update_fields is not None:
            intent_builder = intent_builder.with_update_fields(update_fields)
        intent = intent_builder.build()
        context = OperationContext(intent).with_instances(new_records=tuple(objs))

        return get_orchestrator().execute_create(context, objs)

    def _bulk_update_v3(
        self,
        objs,
        fields,
        batch_size=None,
        bypass_hooks=False,
    ):
        if not objs or not fields:
            return 0

        from django_bulk_hooks.core.context import OperationContext
        from django_bulk_hooks.core.intent import IntentBuilder
        from django_bulk_hooks.orchestration.orchestrator import get_orchestrator

        pks = tuple(obj.pk for obj in objs if obj.pk is not None)
        intent = (
            IntentBuilder(self.model)
            .for_update()
            .targeting(pks)
            .with_update_fields(fields)
            .with_options(
                batch_size=batch_size,
                bypass_hooks=bypass_hooks,
                using_db=self.db,
            )
            .build()
        )
        context = OperationContext(intent).with_instances(new_records=tuple(objs))
        return get_orchestrator().execute_update(context, objs, fields)

    def _update_v3(self, update_kwargs, bypass_hooks=False):
        if not update_kwargs:
            return 0

        from django_bulk_hooks.core.context import OperationContext
        from django_bulk_hooks.core.intent import IntentBuilder
        from django_bulk_hooks.orchestration.orchestrator import get_orchestrator

        pks = tuple(self.values_list("pk", flat=True))
        if not pks:
            return 0

        intent = (
            IntentBuilder(self.model)
            .for_update()
            .targeting(pks)
            .setting(**update_kwargs)
            .with_options(
                bypass_hooks=bypass_hooks,
                using_db=self.db,
            )
            .build()
        )
        context = OperationContext(intent)
        return get_orchestrator().execute_queryset_update(context, self, update_kwargs)

    def _delete_v3(self, bypass_hooks=False):
        from django_bulk_hooks.core.context import OperationContext
        from django_bulk_hooks.core.intent import IntentBuilder
        from django_bulk_hooks.orchestration.orchestrator import get_orchestrator

        pks = tuple(self.values_list("pk", flat=True))
        if not pks:
            return (0, {})

        intent = (
            IntentBuilder(self.model)
            .for_delete()
            .targeting(pks)
            .with_options(
                bypass_hooks=bypass_hooks,
                using_db=self.db,
            )
            .build()
        )
        context = OperationContext(intent)
        return get_orchestrator().execute_delete(context, self)

    def _bulk_delete_v3(self, objs, bypass_hooks=False):
        pks = extract_pks(objs)
        if not pks:
            return 0

        filtered_qs = HookQuerySet.with_hooks(self.filter(pk__in=pks))
        count, _details = filtered_qs._delete_v3(bypass_hooks=bypass_hooks)
        return count



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from types import MappingProxyType
from typing import Any, FrozenSet, Iterable, Mapping, Optional
from typing import Literal
import uuid


@dataclass(frozen=True)
class OperationIntent:
    """
    Immutable description of what the user wants to do.

    This is the single source of truth for the entire operation.
    """

    # Identity
    operation_id: uuid.UUID = field(default_factory=uuid.uuid4)
    created_at: datetime = field(default_factory=datetime.utcnow)

    # What
    operation_type: Literal["create", "update", "delete"] = "update"
    model_cls: type | None = None

    # Target selection
    target_pks: Optional[tuple[Any, ...]] = None

    # Field updates (for update operations)
    field_values: Optional[Mapping[str, Any]] = None
    update_fields: Optional[FrozenSet[str]] = None
    unique_fields: Optional[FrozenSet[str]] = None

    # Execution options
    batch_size: Optional[int] = None
    ignore_conflicts: bool = False
    update_conflicts: bool = False
    bypass_hooks: bool = False

    # Database
    using_db: Optional[str] = None

    def is_field_in_scope(self, field_name: str) -> bool:
        """
        Check if a field is part of this operation's scope.

        If update_fields is None, all fields are in scope.
        """
        if self.update_fields is None:
            return True
        return field_name in self.update_fields

    def get_scoped_fields(self) -> FrozenSet[str]:
        """Return all fields in scope, resolving None to all model fields."""
        if self.update_fields is not None:
            return self.update_fields
        if self.model_cls is None:
            return frozenset()
        return frozenset(
            f.name
            for f in self.model_cls._meta.get_fields()
            if f.concrete and not f.many_to_many
        )


class IntentBuilder:
    """
    Fluent builder for OperationIntent with minimal validation.
    """

    def __init__(self, model_cls: type):
        self.model_cls = model_cls
        self._operation_type: Optional[str] = None
        self._target_pks: Optional[tuple[Any, ...]] = None
        self._field_values: Optional[Mapping[str, Any]] = None
        self._update_fields: Optional[FrozenSet[str]] = None
        self._unique_fields: Optional[FrozenSet[str]] = None
        self._options: dict[str, Any] = {}

    def for_create(self) -> "IntentBuilder":
        self._operation_type = "create"
        return self

    def for_update(self) -> "IntentBuilder":
        self._operation_type = "update"
        return self

    def for_delete(self) -> "IntentBuilder":
        self._operation_type = "delete"
        return self

    def targeting(self, pks: Iterable[Any]) -> "IntentBuilder":
        self._target_pks = tuple(pks)
        return self

    def setting(self, **field_values: Any) -> "IntentBuilder":
        self._field_values = MappingProxyType(dict(field_values))
        self._update_fields = frozenset(field_values.keys())
        return self

    def with_update_fields(self, fields: Iterable[str]) -> "IntentBuilder":
        self._update_fields = frozenset(fields)
        return self

    def with_unique_fields(self, fields: Optional[Iterable[str]]) -> "IntentBuilder":
        if fields is None:
            self._unique_fields = None
            return self
        self._unique_fields = frozenset(fields)
        return self

    def with_options(self, **options: Any) -> "IntentBuilder":
        self._options.update(options)
        return self

    def build(self) -> OperationIntent:
        if self._operation_type is None:
            raise ValueError("Operation type must be specified")
        if self.model_cls is None:
            raise ValueError("Model class must be provided")

        return OperationIntent(
            operation_type=self._operation_type,
            model_cls=self.model_cls,
            target_pks=self._target_pks,
            field_values=self._field_values,
            update_fields=self._update_fields,
            unique_fields=self._unique_fields,
            **self._options,
        )



from __future__ import annotations

import logging
from types import MappingProxyType
from typing import Any, Iterable, Mapping, Optional, Set, Tuple

from django.db.models import Model

from .intent import OperationIntent

logger = logging.getLogger(__name__)


class OperationContext:
    """
    Lazy-loading wrapper around OperationIntent and database state.
    """

    __slots__ = (
        "intent",
        "_old_records_cache",
        "_new_records_cache",
        "_changeset_cache",
        "_select_related",
        "_prefetch_related",
    )

    def __init__(self, intent: OperationIntent):
        self.intent = intent
        self._old_records_cache: Optional[Mapping[Any, Model]] = None
        self._new_records_cache: Optional[Tuple[Model, ...]] = None
        self._changeset_cache = None
        self._select_related: Set[str] = set()
        self._prefetch_related: Set[str] = set()

    # Delegation to intent
    @property
    def operation_type(self) -> str:
        return self.intent.operation_type

    @property
    def model_cls(self) -> type:
        return self.intent.model_cls

    @property
    def operation_id(self):
        return self.intent.operation_id

    def is_field_in_scope(self, field_name: str) -> bool:
        return self.intent.is_field_in_scope(field_name)

    # Lazy data access
    @property
    def old_records_map(self) -> Mapping[Any, Model]:
        if self._old_records_cache is not None:
            logger.debug(
                "[CONTEXT] old_records_map: using cached data (%d records)",
                len(self._old_records_cache),
            )
            return self._old_records_cache

        if self.intent.operation_type == "create":
            logger.debug("[CONTEXT] old_records_map: CREATE operation, returning empty map")
            self._old_records_cache = MappingProxyType({})
            return self._old_records_cache

        if not self.intent.target_pks:
            logger.debug("[CONTEXT] old_records_map: no target_pks, returning empty map")
            self._old_records_cache = MappingProxyType({})
            return self._old_records_cache

        logger.info(
            "[CONTEXT] old_records_map: Loading from DB for model=%s, count=%d PKs",
            self.model_cls.__name__,
            len(self.intent.target_pks),
        )
        qs = self.model_cls._base_manager.filter(pk__in=self.intent.target_pks)
        if self.intent.using_db:
            qs = qs.using(self.intent.using_db)
        if self._select_related:
            qs = qs.select_related(*self._select_related)
        if self._prefetch_related:
            qs = qs.prefetch_related(*self._prefetch_related)

        records = qs.in_bulk(self.intent.target_pks)
        logger.info("[CONTEXT] old_records_map: Loaded %d records from DB", len(records))
        self._old_records_cache = MappingProxyType(records)
        return self._old_records_cache

    @property
    def new_records(self) -> Tuple[Model, ...]:
        if self._new_records_cache is not None:
            logger.debug(
                "[CONTEXT] new_records: using cached data (%d records)",
                len(self._new_records_cache),
            )
            return self._new_records_cache

        if self.intent.operation_type == "delete":
            logger.debug("[CONTEXT] new_records: DELETE operation, using old_records_map")
            self._new_records_cache = tuple(self.old_records_map.values())
            return self._new_records_cache

        if not self.intent.target_pks:
            logger.debug("[CONTEXT] new_records: no target_pks, returning empty tuple")
            self._new_records_cache = tuple()
            return self._new_records_cache

        logger.info(
            "[CONTEXT] new_records: Loading from DB for model=%s, count=%d PKs",
            self.model_cls.__name__,
            len(self.intent.target_pks),
        )
        qs = self.model_cls._base_manager.filter(pk__in=self.intent.target_pks)
        if self.intent.using_db:
            qs = qs.using(self.intent.using_db)
        if self._select_related:
            qs = qs.select_related(*self._select_related)
        if self._prefetch_related:
            qs = qs.prefetch_related(*self._prefetch_related)

        self._new_records_cache = tuple(qs)
        logger.info("[CONTEXT] new_records: Loaded %d records from DB", len(self._new_records_cache))
        return self._new_records_cache

    @property
    def changeset(self):
        if self._changeset_cache is not None:
            return self._changeset_cache

        from .changeset import ChangeSet

        self._changeset_cache = ChangeSet.from_context(self)
        return self._changeset_cache

    # Preloading
    def with_preloaded_relations(self, relations: Iterable[str]) -> "OperationContext":
        relations_set = set(relations)
        if relations_set != self._select_related:
            self._invalidate_caches()
        self._select_related = relations_set
        return self

    def with_prefetched_relations(self, relations: Iterable[str]) -> "OperationContext":
        relations_set = set(relations)
        if relations_set != self._prefetch_related:
            self._invalidate_caches()
        self._prefetch_related = relations_set
        return self

    # Instance injection
    def with_instances(
        self,
        new_records: Optional[Tuple[Model, ...]] = None,
        old_records_map: Optional[Mapping[Any, Model]] = None,
    ) -> "OperationContext":
        if new_records is not None:
            self._new_records_cache = new_records
        if old_records_map is not None:
            self._old_records_cache = MappingProxyType(dict(old_records_map))
        self._changeset_cache = None
        return self

    def clear_caches(self) -> None:
        self._invalidate_caches()

    def _invalidate_caches(self) -> None:
        self._old_records_cache = None
        self._new_records_cache = None
        self._changeset_cache = None



import inspect
import logging

from django_bulk_hooks.core.changeset import RecordChange
from django_bulk_hooks.core.context import OperationContext

logger = logging.getLogger(__name__)


class SignatureAdapter:
    """
    Adapter for hook method signatures.
    """

    @staticmethod
    def call_hook(method, context: OperationContext) -> None:
        sig = inspect.signature(method)
        params = [p for p in sig.parameters.values() if p.name != "self"]
        names = [p.name for p in params]

        if "ctx" in names or "context" in names:
            return method(context)

        changeset = context.changeset
        new_records = changeset.new_records
        old_records = changeset.old_records

        if all(name in {"changeset", "new_records", "old_records"} for name in names):
            positional = []
            for name in names:
                if name == "changeset":
                    positional.append(changeset)
                elif name == "new_records":
                    positional.append(new_records)
                elif name == "old_records":
                    positional.append(old_records)
            return method(*positional)

        kwargs = {}
        if "changeset" in names:
            kwargs["changeset"] = changeset
        if "new_records" in names:
            kwargs["new_records"] = new_records
        if "old_records" in names:
            kwargs["old_records"] = old_records

        if kwargs:
            return method(**kwargs)

        return method()


class ConditionAdapter:
    """
    Adapter for condition signatures.
    """

    @staticmethod
    def check(condition, record: RecordChange, context: OperationContext) -> bool:
        condition_name = condition.__class__.__name__
        
        # Get field name if available for logging
        field_name = getattr(condition, 'field', None)
        pk = record.pk
        
        logger.debug(
            "[CONDITION] Evaluating %s(field=%s) for record pk=%s",
            condition_name,
            field_name,
            pk,
        )
        
        sig = inspect.signature(condition.check)
        if "context" in sig.parameters:
            # New signature with context
            logger.debug("[CONDITION] Using new signature (with context)")
            result = condition.check(record, context)
        else:
            # Old signature without context
            logger.debug("[CONDITION] Using legacy signature (without context)")
            try:
                result = condition.check(record.new_record, record.old_record)
            except TypeError:
                result = condition.check(record.new_record)
        
        # Log detailed information for specific conditions
        if field_name and hasattr(record, 'get_new_value'):
            new_val = record.get_new_value(field_name)
            old_val = record.get_old_value(field_name) if record.old_record else None
            in_scope = context.is_field_in_scope(field_name)
            
            logger.info(
                "[CONDITION] %s(field=%s): pk=%s, result=%s, in_scope=%s, old=%s, new=%s",
                condition_name,
                field_name,
                pk,
                result,
                in_scope,
                old_val,
                new_val,
            )
        else:
            logger.debug("[CONDITION] %s: pk=%s, result=%s", condition_name, pk, result)
        
        return result



import logging

from django_bulk_hooks.enums import DEFAULT_PRIORITY
from django_bulk_hooks.registry import register_hook, unregister_hook

from .requirements import HookRequirements

logger = logging.getLogger(__name__)


class HookMeta(type):
    _registered = set()
    _class_hook_map: dict[type, set[tuple]] = {}

    def __new__(mcs, name, bases, namespace):
        cls = super().__new__(mcs, name, bases, namespace)
        mcs._register_hooks_for_class(cls)
        return cls

    @classmethod
    def _register_hooks_for_class(mcs, cls):
        # Unregister parent hooks to ensure overrides win
        for base in cls.__mro__[1:]:
            if not isinstance(base, HookMeta):
                continue
            if base in mcs._class_hook_map:
                for model_cls, event, base_cls, method_name in list(
                    mcs._class_hook_map[base],
                ):
                    key = (model_cls, event, base_cls, method_name)
                    if key in HookMeta._registered:
                        unregister_hook(model_cls, event, base_cls, method_name)
                        HookMeta._registered.discard(key)

        all_hook_methods = {}
        for klass in reversed(cls.__mro__):
            if not isinstance(klass, HookMeta):
                continue
            for method_name, method in klass.__dict__.items():
                if hasattr(method, "_hook_registrations"):
                    all_hook_methods[method_name] = method

        if cls not in mcs._class_hook_map:
            mcs._class_hook_map[cls] = set()

        for method_name, method in all_hook_methods.items():
            for reg in getattr(method, "_hook_registrations", []):
                model_cls = reg["model"]
                event = reg["event"]
                condition = reg.get("condition")
                priority = reg.get("priority", DEFAULT_PRIORITY)
                key = (model_cls, event, cls, method_name)
                if key in HookMeta._registered:
                    continue
                register_hook(
                    model=model_cls,
                    event=event,
                    handler_cls=cls,
                    method_name=method_name,
                    condition=condition,
                    priority=priority,
                )
                HookMeta._registered.add(key)
                mcs._class_hook_map[cls].add(key)


class Hook(metaclass=HookMeta):
    """
    Base class for hook handlers with declarative requirements.
    """

    requirements: HookRequirements = HookRequirements.standard()

    def get_requirements(self, event: str) -> HookRequirements:
        return self.requirements



from dataclasses import dataclass, field
from typing import FrozenSet, Optional


@dataclass(frozen=True)
class HookRequirements:
    """
    Declarative requirements for a hook handler.
    """

    old_records: bool = False
    new_records: bool = True
    preload_related: FrozenSet[str] = field(default_factory=frozenset)
    prefetch_related: FrozenSet[str] = field(default_factory=frozenset)
    batch_aware: bool = False
    max_batch_size: Optional[int] = None

    @classmethod
    def none(cls) -> "HookRequirements":
        return cls(old_records=False, new_records=False)

    @classmethod
    def standard(cls) -> "HookRequirements":
        return cls(old_records=False, new_records=True)

    @classmethod
    def full(cls) -> "HookRequirements":
        return cls(old_records=True, new_records=True)



import logging
from typing import List

from django.db.models import Model, QuerySet

from django_bulk_hooks.core.context import OperationContext
from django_bulk_hooks.operations.field_utils import handle_auto_now_fields

logger = logging.getLogger(__name__)


class DatabaseExecutor:
    """
    Execute database operations with no hook logic.
    """

    def bulk_create(self, context: OperationContext, objs: List[Model]) -> List[Model]:
        intent = context.intent
        qs = intent.model_cls._base_manager.all()
        if intent.using_db:
            qs = qs.using(intent.using_db)

        update_fields = list(intent.update_fields) if intent.update_fields else None
        if intent.update_conflicts and update_fields:
            update_fields = self._remove_auto_now_add_fields(update_fields, objs)

        return qs.bulk_create(
            objs,
            batch_size=intent.batch_size,
            ignore_conflicts=intent.ignore_conflicts,
            update_conflicts=intent.update_conflicts,
            update_fields=update_fields,
            unique_fields=list(intent.unique_fields) if intent.unique_fields else None,
        )

    def bulk_update(
        self,
        context: OperationContext,
        objs: List[Model],
        fields: List[str],
    ) -> int:
        intent = context.intent
        qs = intent.model_cls._base_manager.all()
        if intent.using_db:
            qs = qs.using(intent.using_db)
        if not objs:
            return 0

        fields = self._add_auto_now_fields(fields, objs)
        fields = self._remove_auto_now_add_fields(fields, objs)
        return qs.bulk_update(objs, fields, batch_size=intent.batch_size)

    def queryset_update(self, queryset: QuerySet, **kwargs) -> int:
        return QuerySet.update(queryset, **kwargs)

    def queryset_delete(self, queryset: QuerySet):
        return QuerySet.delete(queryset)

    def _add_auto_now_fields(self, fields: List[str], objs: List[Model]) -> List[str]:
        if not objs:
            return fields
        auto_now_fields = handle_auto_now_fields(objs[0].__class__, objs, for_update=True)
        for field_name in auto_now_fields:
            if field_name not in fields:
                fields.append(field_name)
        return fields

    def _remove_auto_now_add_fields(self, fields: List[str], objs: List[Model]) -> List[str]:
        if not objs:
            return fields
        model_cls = objs[0].__class__
        auto_now_add_fields = set()
        for field in model_cls._meta.get_fields():
            if getattr(field, "auto_now_add", False):
                auto_now_add_fields.add(field.name)
        if not auto_now_add_fields:
            return fields
        filtered = [field for field in fields if field not in auto_now_add_fields]
        removed = auto_now_add_fields.intersection(fields)
        if removed:
            logger.info(
                "Removed auto_now_add fields from update fields list: %s",
                ", ".join(sorted(removed)),
            )
        return filtered



from __future__ import annotations

import logging
from typing import Any, FrozenSet, Iterable, Optional, Tuple

from django.db.models import Model

from .context import OperationContext

logger = logging.getLogger(__name__)


class RecordChange:
    """
    Single record change with context awareness.
    """

    __slots__ = ("_new", "_old", "_context", "_changed_fields_cache")

    def __init__(
        self,
        new_record: Model,
        old_record: Optional[Model],
        context: Optional[OperationContext] = None,
    ):
        self._new = new_record
        self._old = old_record
        self._context = context
        self._changed_fields_cache: Optional[FrozenSet[str]] = None

    @property
    def new_record(self) -> Model:
        return self._new

    @property
    def old_record(self) -> Optional[Model]:
        return self._old

    @property
    def pk(self) -> Any:
        return self._new.pk if self._new else None

    @property
    def context(self) -> Optional[OperationContext]:
        return self._context

    def has_changed(self, field_name: str) -> bool:
        if self._context is not None and not self._context.is_field_in_scope(field_name):
            return False
        if self._old is None:
            return False
        return field_name in self.changed_fields

    @property
    def changed_fields(self) -> FrozenSet[str]:
        if self._changed_fields_cache is not None:
            return self._changed_fields_cache
        if self._old is None:
            self._changed_fields_cache = frozenset()
            return self._changed_fields_cache

        from django_bulk_hooks.operations.field_utils import get_changed_fields

        all_changed = get_changed_fields(self._old, self._new, self._new.__class__)
        if self._context is not None and self._context.intent.update_fields is not None:
            scoped = all_changed & self._context.intent.update_fields
            logger.debug(
                "[RECORD_CHANGE] pk=%s: all_changed=%s, update_fields=%s, scoped=%s",
                self.pk,
                all_changed,
                self._context.intent.update_fields,
                scoped,
            )
            self._changed_fields_cache = frozenset(scoped)
        else:
            logger.debug(
                "[RECORD_CHANGE] pk=%s: all_changed=%s (no field scope filtering)",
                self.pk,
                all_changed,
            )
            self._changed_fields_cache = frozenset(all_changed)
        return self._changed_fields_cache

    def get_old_value(self, field_name: str) -> Any:
        if self._old is None:
            return None
        return getattr(self._old, field_name, None)

    def get_new_value(self, field_name: str) -> Any:
        return getattr(self._new, field_name, None)


class ChangeSet:
    """
    Immutable view over operation changes.
    """

    __slots__ = ("_context", "_changes", "_pk_index")

    def __init__(self, context: OperationContext, changes: Tuple[RecordChange, ...]):
        self._context = context
        self._changes = changes
        self._pk_index = {c.pk: c for c in changes if c.pk is not None}

    @classmethod
    def from_context(cls, context: OperationContext) -> "ChangeSet":
        logger.debug(
            "[CHANGESET] Building ChangeSet for model=%s, operation=%s",
            context.model_cls.__name__,
            context.operation_type,
        )
        new_records = context.new_records
        old_records_map = context.old_records_map
        
        logger.info(
            "[CHANGESET] Creating changes: new_records=%d, old_records=%d",
            len(new_records),
            len(old_records_map),
        )
        
        changes = tuple(
            RecordChange(
                new_record=new,
                old_record=old_records_map.get(new.pk),
                context=context,
            )
            for new in new_records
        )
        
        logger.debug("[CHANGESET] ChangeSet created with %d changes", len(changes))
        return cls(context, changes)

    @property
    def context(self) -> OperationContext:
        return self._context

    @property
    def intent(self):
        return self._context.intent

    @property
    def operation_type(self) -> str:
        return self._context.operation_type

    @property
    def model_cls(self) -> type:
        return self._context.model_cls

    @property
    def new_records(self) -> list[Model]:
        return [c.new_record for c in self._changes]

    @property
    def old_records(self) -> list[Model]:
        return [c.old_record for c in self._changes if c.old_record is not None]

    @property
    def changes(self) -> Tuple[RecordChange, ...]:
        return self._changes

    def get_change(self, pk: Any) -> Optional[RecordChange]:
        return self._pk_index.get(pk)

    def has_field_changed(self, pk: Any, field_name: str) -> bool:
        change = self._pk_index.get(pk)
        return change.has_changed(field_name) if change else False

    def __len__(self) -> int:
        return len(self._changes)

    def __iter__(self):
        return iter(self._changes)

    def chunk(self, chunk_size: int) -> Iterable["ChangeSet"]:
        for i in range(0, len(self._changes), chunk_size):
            chunk_changes = self._changes[i : i + chunk_size]
            yield ChangeSet(self._context, chunk_changes)
